-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local im  = ui_imgui

local C = {}

C.name = 'File Race for Extra AI race mission'
C.description = 'Manages a Race from a given path for extra AI race mission.'
-- C.category = 'repeat_instant'
C.color = im.ImVec4(1, 1, 0, 0.75)
C.pinSchema = {
  { dir = 'in', type = 'flow', name = 'flow', description = 'Inflow for this node.' },
  { dir = 'in', type = 'flow', name = 'idle', hidden = true, description = 'If this pin has flow, the race will not count the timer and not count as active.' },
  { dir = 'in', type = 'flow', name = 'reset', description = 'Resets this node.', impulse = true },
  { dir = 'in', type = 'number', name = 'lapCount', default = 1, hardcoded = true, hidden = true, description = 'Number of laps (min 1).' },
  { dir = 'in', type = 'table', name = 'pathData', tableType = 'pathData', description = 'Path data' },
  { dir = 'in', type = 'bool', name = 'rolling', description = 'If the path should be with a rolling start, if possible.'},
  {dir = 'in', type = 'number', name = 'playerId', description = 'Player vehicle ID.'},
  {dir = 'in', type = 'bool', name = 'watchingAIMode', description = 'Use watching AI mode or not.'},

  { dir = 'out', type = 'flow', name = 'flow', description = 'Outflow from this node.' },
  { dir = 'out', type = 'table', name = 'raceData', tableType = 'raceData',  description = 'Data from the race for other nodes to process.' },
  { dir = 'out', type = 'table', name = 'aiPath', tableType = 'navgraphPath',  description = 'AI navgraph path; can be used with the Follow Waypoints node.' },
  { dir = 'out', type = 'flow', name = 'active', description = 'Outflow when race is active.' },
  { dir = 'out', type = 'flow', name = 'complete', description = 'Outflow when race is complete.' },
  { dir = 'out', type = 'number', name = 'time', description = 'Total time when completed.' }
}

C.tags = {'scenario'}

function C:init(mgr, ...)
  self.started = false
  self.race = nil
  self.data.useHotlappingApp = true
  self.data.useDebugDraw = false
  self.data.useWaypointAudio = true
end

function C:_executionStarted()
  self.race = nil
  self.started = false
end

function C:_executionStopped()
  if self.race then self.race:stopRace() end
  self.race = nil
  self.started = false
end

function C:drawMiddle(builder, style)
  builder:Middle()
  im.Text(self.started and "Started" or "Stopped")
end

function C:work(args)
  if self.pinIn.reset.value then
    self.started = false
    self.pinOut.flow.value = false
    self.pinOut.complete.value = false
  end
  if self.pinIn.flow.value then
    if not self.started then
      self.race = require('/lua/ge/extensions/gameplay/race/race')()
      self.race:setPath(self.pinIn.pathData.value)
      self.race.path.config.rollingStart = self.pinIn.rolling.value
      self.race.lapCount = self.pinIn.pathData.value.config.closed and math.max(self.pinIn.lapCount.value or 1, 1) or 1
      self.race.useHotlappingApp = self.data.useHotlappingApp
      self.race.useDebugDraw = self.data.useDebugDraw
      self.race.useWaypointAudio = self.data.useWaypointAudio

      if self.mgr.activity then
        self.race.saveFileSuffix = self.mgr.activity.id
      end
      local vids = {}
      if not self.pinIn.watchingAIMode.value then
        table.insert(vids, self.pinIn.playerId.value)
      end
      for vid, _ in pairs(mathkuro_extraRaceMission.getRacerInfoTable()) do
        if vid then
          table.insert(vids, vid)
        end
      end
      self.race:setVehicleIds(vids)
      self.race:startRace()
      self.started = true
    end
    if not self.race then return end
    if not self.pinIn.idle.value then
      self.race:onUpdate(self.mgr.dtSim)
    end
    self.pinOut.raceData.value = self.race
    self.pinOut.aiPath.value = self.race.aiPath
    self.pinOut.flow.value = self.pinIn.flow.value

    local state = self.race.states[self.pinIn.playerId.value]
    self.pinOut.complete.value = state and state.complete or false
    if self.pinOut.complete.value then
      self.pinOut.time.value = state.historicTimes[#state.historicTimes].endTime
    end
    self.pinOut.active.value = state and state.active or false
    if not self.pinOut.active.value and self.pinIn.idle.value ~= nil then
      self.pinOut.active.value = not self.pinIn.idle.value
    end
  end
end


return _flowgraph_createNode(C)
