local M = {}
local logTag = 'mathkuro'

local DAMAGE_TBL_KEY = 10001
local DAMAGE_BY_VAL_TBL_KEY = 10002

local TOTAL_DAMAGE_BY_VAL_TBL_KEY = 20001
local TOTAL_DAMAGED_VEHICLE_NUM_TBL_KEY = 20002

-- ダメージ値を$ベースに換算する係数(要チューニング)
local DAMAGE_TO_VALUE = 25000

-- value未定義の車両のvalue値
local DEFAULT_VEHICLE_VALUE = 30000

-- ダメージ計算を行うダメージ[$]の閾値
local DAMAGE_CALC_THRESHOLD = 25

-- ダメージテーブル保持期間(プレイヤーとクラッシュして一定時間のダメージをダメージとして計上する)
local DAMAGE_TABLE_REMAIN_FRAME_THRESHOLD = 120
local FORCE_FIELD_DAMAGE_TABLE_REMAIN_FRAME_THRESHOLD = 300
local damageTableRemainFrameThreshold = DAMAGE_TABLE_REMAIN_FRAME_THRESHOLD

-- クラッシュカメラの設定
local SIM_SPEED = 0.2
local CRASH_CAMERA_MODE = 'external'

-- カスタム車両のファイルパス
local VEHICLE_FILE_DIR = '/gameplay/missionTypes/mathkuroMission02/rivalVehGroup/'
local BASE_VEHICLE_FILE_NAME = 'base.vehGroup.json'
local CUSTOM_VEHICLE_FILE_NAME = 'custom.vehGroup.json'

local UI_SETTINGS_PATH = '/ui/modules/apps/extraRaceMissionMathkuro/settings.json'

-- グローバル変数
local racerInfoTable = {}

-- クラッシュ系
local currentDamageTable = {}
local lastPlayerDamage = -1
local lastCameraName = ''
local isCrashCameraMode = false

local totalDamageValue = 0
local totalDamagedVehicleNum = 0

-- 被害総額算出対象とする車両距離
local DEFAULT_DISTANCE_THRESHOLD = 5
local FORCE_FIELD_DISTANCE_THRESHOLD = 100
local distanceThreshold = DEFAULT_DISTANCE_THRESHOLD

-- 順位系
local currentRacePosition = 0
local startPositionTable = {}

-- 追跡情報系
local _lastOffensesTable = {}
local _lastHitCount = 0
local _isSpeeding = false
local _isRacing = false
local _isReckless = false
local _isWrongWay = false
local _isIntersection = false

-- UI設定
local enableCheckPoint = true
local enableCheckPointSound = false
local displayGroundMarker = false

local playerResetState = false
local function setPlayerResetState(state)
  playerResetState = state
  -- print('setCurrentRacePosition'..tostring(currentRacePosition))
end
local function getPlayerResetState()
  -- print('getCurrentRacePosition'..tostring(currentRacePosition))
  return playerResetState
end

local function setCurrentRacePosition(position)
  currentRacePosition = position
  -- print('setCurrentRacePosition'..tostring(currentRacePosition))
end
local function getCurrentRacePosition()
  -- print('getCurrentRacePosition'..tostring(currentRacePosition))
  return currentRacePosition
end

local function getTable(key)
  if not currentDamageTable[key] then
    currentDamageTable[key] = {}
  end

  return currentDamageTable[key]
end

local function setDistanceThreshold(threshold)
  distanceThreshold = threshold
end

local function resetDamageValues(key)
  local _key = (type(key) == 'number') and key or DAMAGE_BY_VAL_TBL_KEY
  currentDamageTable[_key] = {}
  totalDamageValue = 0
  totalDamagedVehicleNum = 0
  return
end

local function getTotalDamageValue()
  return math.floor(totalDamageValue)
end

local function getDamagedVehicleNum()
  return totalDamagedVehicleNum
end

-- プレイヤーのダメージ差分(前回呼び出し時との差分)を取得
local function getPlayerDamageDelta()
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil or map.objects[_playerVeh:getId()] == nil then
    return 0
  end

  local _playerDamageDelta = 0
  local _playerDamage = map.objects[_playerVeh:getId()]["damage"]

  if _playerDamage < lastPlayerDamage then
    lastPlayerDamage = 0
  end

  if lastPlayerDamage == -1 then
    lastPlayerDamage = _playerDamage
    -- 初回呼び出し時に_playerDamageの値に関わらず差分は0として扱う
    _playerDamageDelta = 0
  else
    _playerDamageDelta = _playerDamage - lastPlayerDamage
    lastPlayerDamage = _playerDamage
  end

  return _playerDamageDelta
end

-- 車両毎のダメージテーブルを更新
local function updateDamagedVehicleTable(key)
  local _key = (type(key) == 'number') and key or DAMAGE_BY_VAL_TBL_KEY
  local _tbl = getTable(_key)
  local _totalDamageDiff = 0
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil then
    return
  end

  local _playerVid = _playerVeh:getId()
  local _playerPos = _playerVeh:getPosition()
  local _playerLenX = be:getObjectByID(_playerVid).initialNodePosBB:getExtents().x
  local _playerLenY = be:getObjectByID(_playerVid).initialNodePosBB:getExtents().y
  local _playerCurrentDamage = getPlayerDamageDelta()

  if extensions.gameplay_forceField.isActive() then
    setDistanceThreshold(FORCE_FIELD_DISTANCE_THRESHOLD)
  end

  for vid, veh in activeVehiclesIterator() do
    local _currentDamage = 0
    -- 前回のダメージ値(初出の車両の場合は0)
    local _lastDamage = 0
    if _tbl[vid] and _tbl[vid]['damage'] then
      _lastDamage = _tbl[vid]['damage']
    end

    if map.objects[vid] ~= nil then
      _currentDamage = map.objects[vid]['damage']
    end

    if _currentDamage < _lastDamage or _currentDamage < DAMAGE_CALC_THRESHOLD then
      -- 以下の場合はテーブルから削除
      -- 前回値よりもダメージ量が小さい場合(車両リセット)
      -- DAMAGE_CALC_THRESHOLD以下(表示数削減)
      _tbl[vid] = nil
    else
      if vid ~= _playerVid then
        local _distance = veh:getPosition():distance(_playerPos)
        local _vehLenX = be:getObjectByID(vid).initialNodePosBB:getExtents().x
        local _vehLenY = be:getObjectByID(vid).initialNodePosBB:getExtents().y
        local _nearestDistX = math.min(_playerLenX * 0.5 + _vehLenX * 0.5, _distance)
        local _nearestDistY = math.min(_playerLenY * 0.5 + _vehLenY * 0.5, _distance)
        local _squareDistance = veh:getPosition():squaredDistance(_playerPos)
        if (_tbl[vid] or _squareDistance <= square(_nearestDistX) or (_squareDistance <= square(_nearestDistY) and 0 < _playerCurrentDamage)) then
          -- 登録済みまたはプレイヤー車両との距離が閾値未満の車両のみダメージ値更新

          if _tbl[vid] == nil or _tbl[vid]['remain_frame'] == nil then
            -- log('D', logTag, '_distance:'.._distance..', _playerLenX:'.._playerLenX..', _vehLenX:'.._vehLenX..', _nearestDistX:'.._nearestDistX..', _nearestDistY:'.._nearestDistY..', _squareDistance:'.._squareDistance..', square(_nearestDistX):'..square(_nearestDistX))
            -- 新規登録車両は破壊台数に1加算
            totalDamagedVehicleNum = totalDamagedVehicleNum + 1

            _tbl[vid] = {}

            -- 車両名取得
            if string.find(core_vehicles.getVehicleDetails(vid).configs.Name, 'simple_traffic') == nil then
              _tbl[vid]['vname'] = core_vehicles.getVehicleDetails(vid).configs.Name
            else
              _tbl[vid]['vname'] = core_vehicles.getVehicleDetails(vid).model.Brand.." "..core_vehicles.getVehicleDetails(vid).model.Name
            end

            -- 車両価格
            local _value = core_vehicles.getVehicleDetails(vid).configs.Value
            if _value == nil then
              -- Value値が未設定の車両がたまにあるので補正
              _value = DEFAULT_VEHICLE_VALUE * (1 + math.random())
            end
            _tbl[vid]['value'] = _value

            -- 各種初期値設定
            _tbl[vid]['remain_frame'] = 0
            _tbl[vid]['damage'] = 0
          end

          if _tbl[vid]['remain_frame'] < DAMAGE_TABLE_REMAIN_FRAME_THRESHOLD then
            -- プレイヤー車両との距離が閾値未満になってから一定期間のみダメージ計算を行う
            _tbl[vid]['damage'] = _currentDamage
            _tbl[vid]['remain_frame'] = _tbl[vid]['remain_frame'] + 1
            -- log('D', logTag, 'remain_frame:'.._tbl[vid]['remain_frame'])

            -- ダメージ差分
            -- 車両価格の2倍まで許容？
            local _damageRate = ((_currentDamage - _lastDamage) / DAMAGE_TO_VALUE)
            _damageRate = (2.0 > _damageRate) and _damageRate or 2.0
            _totalDamageDiff = _totalDamageDiff + (_damageRate * _tbl[vid]['value'])
          end

          if _squareDistance <= square(_nearestDistY) then
            _tbl[vid]['remain_frame'] = 0
          end
        end
      end
    end
  end

  totalDamageValue = totalDamageValue + _totalDamageDiff
end

-- 車両毎のダメージテーブルを取得
local function getDamagedVehicleTable(key)
  local _key = (type(key) == 'number') and key or DAMAGE_BY_VAL_TBL_KEY
  return getTable(_key)
end

-- from lua\ge\extensions\gameplay\traffic\vehicle.lua C:checkOffenses()
local function checkOffenses() -- tests for vechicle offenses for police
  local _playerTrafficData = gameplay_traffic.getTrafficData()[be:getPlayerVehicleID(0)]
  if not _playerTrafficData then return end -- if not traffic activated

  local _speed = _playerTrafficData.speed
  local _tracking = _playerTrafficData.tracking
  local _collisions = _playerTrafficData.collisions

  local pursuit = _playerTrafficData.pursuit
  if not pursuit.mode or pursuit.mode < 1 then _lastHitCount = 0 return end -- if not pursuit started

  local minScore = clamp(_playerTrafficData.policeVars.strictness, 0, 0.8) -- offense threshold
  if _playerTrafficData.policeVars.strictness <= 0 then return end

  if _speed >= _tracking.speedLimit * 1.2 and not _isSpeeding then
    _lastOffensesTable['speeding'] = {value = _speed, maxLimit = _tracking.speedLimit, score = 100}
    _isSpeeding = true
  elseif  _speed < _tracking.speedLimit * 1.2 then
    _isSpeeding = false
  end
  if _speed >= _tracking.speedLimit * 2 and not _isRacing then
    _lastOffensesTable['racing'] = {value = _speed, maxLimit = _tracking.speedLimit, score = 200}
    _isRacing = true
  elseif _speed < _tracking.speedLimit * 2 then
    _isRacing = false
  end
  if _tracking.driveScore <= minScore and not _isReckless then
    _lastOffensesTable['reckless'] = {value = _tracking.driveScore, minLimit = minScore, score = 200}
    _isReckless = true
  elseif _tracking.driveScore > minScore then
    _isReckless = false
  end
  if (_tracking.signalFault and not pursuit.offenses.intersection) then
    if not _isIntersection then
      _lastOffensesTable['intersection'] = {value = _tracking.intersectionScore, minLimit = minScore, score = 200}
      _isIntersection = true
    end
  else
    _isIntersection = false
  end
  if _tracking.directionScore <= minScore and not _isWrongWay then
    _lastOffensesTable['wrongWay'] = {value = _tracking.directionScore, minLimit = minScore, score = 150}
    _isWrongWay = true
  elseif _tracking.directionScore > minScore then
    _isWrongWay = false
  end

  if _lastHitCount < pursuit.hitCount then
    for id, coll in pairs(_collisions) do
      if coll.dot >= 0.2 then -- simple comparison to check if current vehicle is at fault for collision
        local veh = gameplay_traffic.getTrafficData()[id]

        _lastHitCount = _lastHitCount + 1
        if veh.role.name == 'police' and coll.inArea then -- always triggers if police was hit
          local score = veh.isPlayerControlled and 50 or 150
          _lastOffensesTable['hitPolice'] = {value = id, score = score}
        elseif pursuit.mode > 0 or coll.state == 'abandoned' then -- fleeing in a pursuit, or abandoning an accident
          _lastOffensesTable['hitTraffic'] = {value = id, score = 100}
        end
      end
    end
  end

end

local function getOffenses()
  local _offensesTable = _lastOffensesTable
  _lastOffensesTable = {}
  return _offensesTable
end

local function resetOffenses()
  _isSpeeding = false
  _isRacing = false
  _isWrongWay = false
end

local function enableCrashCamera()
  -- 現在設定されているカメラを退避
  if isCrashCameraMode then
    -- クラッシュカメラ中はクラッシュカメラを起動しない
    log('D', logTag, 'crashCameraMode already enabled')
  else
    lastCameraName = core_camera.getActiveCamName()
    log('D', logTag, 'lastCameraName:'..lastCameraName)

    bullettime.set(SIM_SPEED)

    local _playerVelocity = be:getPlayerVehicle(0):getVelocity():length()
    log('D', logTag, '_playerVelocity:'.._playerVelocity)

    if _playerVelocity > 15 then
      core_camera.setByName(0, CRASH_CAMERA_MODE)
      core_camera.resetCamera(0)
    else
      core_camera.setByName(0, 'orbit')
      core_camera.cameraZoom(0.002)
      core_camera.rotate_yaw(0.1)
      local _vid = be:getPlayerVehicle(0):getId()
      -- core_camera.setFOV(_vid, 90)
    end

    isCrashCameraMode = true
  end
end

local function disableCrashCamera()
  if isCrashCameraMode then
    -- クラッシュカメラが有効化されている場合のみ解除処理を行う
    bullettime.set(1)
    core_camera.setByName(0, lastCameraName)
    core_camera.resetCamera(0)
    isCrashCameraMode = false
  else
    log('D', logTag, 'crashCameraMode is not enabled')
  end
end

local function initializeRacerInfo(vid, startPositionName, pos, rot)
  log('D', logTag, 'initializeRacerInfo vid=' .. tostring(vid))
  if racerInfoTable[vid] == nil then
    racerInfoTable[vid] = {}
  end

  racerInfoTable[vid] = {
    startPositionName = startPositionName,
    pos = pos,
    rot = rot
  }
end

local function getRacerInfoTable()
  return racerInfoTable
end

local function updateRacerInfo(vid, curCP, curLap, aiAggression, aiPath)
  local racerInfo = racerInfoTable[vid]

  if curCP then
    racerInfo.curCP = curCP
  end
  if curLap then
    racerInfo.curLap = curLap
  end
  if aiAggression then
    racerInfo.aiAggression = aiAggression
  end
  if aiPath then
    racerInfo.aiPath = aiPath
  end
end

local function setAllRacerAIMode(mode)
  log('D', logTag, 'setAllRacerAIMode(' .. tostring(mode) .. ')')
  for vid, _ in pairs(racerInfoTable) do
    local veh = be:getObjectByID(vid)
    veh:queueLuaCommand('ai.setMode("' .. tostring(mode) .. '")')
  end
end

local function revUpAllRacer()
  log('D', logTag, 'revUpAllRacer()')
  for vid, _ in pairs(racerInfoTable) do
    local veh = be:getObjectByID(vid)
    veh:queueLuaCommand('controller.setFreeze(1)')
    veh:queueLuaCommand('ai.setMode("random")')
  end
end

local function resetAllStartPosition()
  log('D', logTag, 'resetAllStartPosition()')
  for vid, racerInfo in pairs(racerInfoTable) do
    log('D', logTag, 'vid=' .. tostring(vid))
    local pos = racerInfo.pos
    -- lua\ge\spawn.lua で謎に180度補正されているので事前に180度ずらしておく
    local rot = quat(racerInfo.rot.x, racerInfo.rot.y, racerInfo.rot.z, racerInfo.rot.w) * quat(0, 0, 1, 0)
    vehicleSetPositionRotation(vid, pos.x, pos.y, pos.z, rot.x, rot.y, rot.z, rot.w)
  end
end

local function getPlayerCurrentDamage()
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil or map.objects[_playerVeh:getId()] == nil then
    return 0
  end

  return map.objects[_playerVeh:getId()]["damage"]
end

local function readUISettings()
  local _settings = jsonReadFile(UI_SETTINGS_PATH)
  if _settings then
    enableCheckPoint = _settings.enableCheckPoint.enabled
    enableCheckPointSound = _settings.enableCheckPointSound.enabled
    displayGroundMarker = _settings.displayGroundMarker.enabled
  end
end

local updateUISettings = function(settings)
  enableCheckPoint = settings.enableCheckPoint.enabled
  enableCheckPointSound = settings.enableCheckPointSound.enabled
  displayGroundMarker = settings.displayGroundMarker.enabled
end

local getEnableCheckPoint = function()
  return enableCheckPoint
end

local getEnableCheckPointSound = function()
  return enableCheckPointSound
end

local getDisplayGroundMarker = function()
  return displayGroundMarker
end

local function resetAll(force)
  log('D', logTag, 'resetAll()')
  if force then
    racerInfoTable = {}
  end

  currentDamageTable = {}
  totalDamageValue = 0
  totalDamagedVehicleNum = 0
  lastPlayerDamage = -1
  distanceThreshold = DEFAULT_DISTANCE_THRESHOLD
  damageTableRemainFrameThreshold = DAMAGE_TABLE_REMAIN_FRAME_THRESHOLD

  currentRacePosition = 0

  _lastOffensesTable = {}
  _lastHitCount = 0

  readUISettings()
end

-- custom.vehGroup.jsonのフォルダを開く
local function openCustomVehicleFolder()
  local customVehicleFile = VEHICLE_FILE_DIR .. CUSTOM_VEHICLE_FILE_NAME
  if not jsonReadFile(customVehicleFile) then
    -- 初回等はファイルが存在しないので、MODのzipからコピーし新規作成
    log('I', logTag, customVehicleFile .. ' not found : create new file')
    local _vehGroup = jsonReadFile(VEHICLE_FILE_DIR .. BASE_VEHICLE_FILE_NAME)
    if _vehGroup then
      jsonWriteFile(customVehicleFile, _vehGroup, true)
    else
      log('E', logTag, 'Base vehicle group file is not found. somthing wrong.')
    end
  end

  Engine.Platform.exploreFolder(VEHICLE_FILE_DIR)
end


M.updateUISettings = updateUISettings

M.getEnableCheckPoint = getEnableCheckPoint
M.getEnableCheckPointSound = getEnableCheckPointSound
M.getDisplayGroundMarker = getDisplayGroundMarker

M.initializeRacerInfo = initializeRacerInfo
M.getRacerInfoTable = getRacerInfoTable
M.updateRacerInfo = updateRacerInfo
M.setAllRacerAIMode = setAllRacerAIMode
M.revUpAllRacer = revUpAllRacer

M.setPlayerResetState = setPlayerResetState
M.getPlayerResetState = getPlayerResetState

M.resetAll = resetAll

M.resetAllStartPosition = resetAllStartPosition

M.getPlayerDamageDelta = getPlayerDamageDelta
M.getPlayerCurrentDamage = getPlayerCurrentDamage

M.enableCrashCamera = enableCrashCamera
M.disableCrashCamera = disableCrashCamera

M.checkOffenses = checkOffenses
M.getOffenses = getOffenses
M.resetOffenses = resetOffenses

M.setCurrentRacePosition = setCurrentRacePosition
M.getCurrentRacePosition = getCurrentRacePosition

M.resetDamageValues = resetDamageValues

M.setDistanceThreshold = setDistanceThreshold

M.getTotalDamageValue = getTotalDamageValue

M.getDamagedVehicleNum = getDamagedVehicleNum
M.updateDamagedVehicleTable = updateDamagedVehicleTable

M.getDamagedVehicleTable = getDamagedVehicleTable

M.openCustomVehicleFolder = openCustomVehicleFolder

return M
